import { BehaviorSubject, Observable, Subject } from "rxjs"
import { RetransmissionService } from "../utils/retransmission.service"
import { WrappedMessage } from "../utils/message.ordering"

/* EVENT BUS */
export interface GeneralEvent {
    id: string,
    event: EventType,
    date: Date
    message?: string
}

export interface TransmissionEvent extends GeneralEvent {
    transmission: Transmission
}

export interface AdapterEvent extends GeneralEvent {
    adapters: AdapterBase | AdapterBase[]
}

export interface TransportEvent extends GeneralEvent {
    data: any
}

/* MANAGEMENT */
export interface MessageTransmissionManager {
    subscribe(): Observable<TransmissionEvent>
}

export interface AdapterManager {
    subscribe(): Observable<AdapterEvent>
}


/* TRANSMISSION COMPONENTS */
export interface MessageTransmissionBase {
    adapters: AdapterBase[]
    mainAdapter: AdapterBase

    getInfo(): TransmissionProfile
    setUpAdapter(adapter: AdapterBase): void
}

export interface MessageReceiver extends MessageTransmissionBase {
    receiverProfile: ReceiverProfile

    getMessageBus(bus: Bus): Observable<any>
    setReceiver(receiverProfile: ReceiverProfile, event: Observable<TransportEvent>): void
}

export interface MessageTransmitter extends MessageTransmissionBase {
    transmitterProfile: TransmitterProfile
    retransmission: RetransmissionService

    setTransmitter(transmitterProfile: TransmitterProfile, event: Observable<TransportEvent>): void
}

export interface MessageRequestResponse extends MessageTransmissionBase {
    transmitterInstance: MessageTransmitter
    receiverInstance: MessageReceiver

    setTransmissionProfile(transmissionInfo: MessageTransmitter, receiverInfo: MessageReceiver): void
}


export interface AdapterProfile {
    id: string,
    transportType: Transport,
}




export interface AdapterBase {
    connector: TransportService // this one will refer to the actual tranpsort service like websocket and so on
    adapterProfile: AdapterProfile
    event: Subject<TransportEvent>

    getInfo(): AdapterProfile
    setAdapterProfile(id: string, transportType: Transport): void
}


export interface TransmitterAdapter extends AdapterBase {
    emit(message: WrappedMessage): void
}

export interface ReceiverAdapter extends AdapterBase {
    getMessageBus(bus: Bus): Observable<any>
}

/* Utils */
export interface TransmissionProfile {
    id: string,
    name: string,
    dateCreated: Date
}

export interface TransmitterProfile extends TransmissionProfile {
}

export interface ReceiverProfile extends TransmissionProfile {
}

export interface RequestResponseProfile extends TransmissionProfile {

}

export enum Bus {
    GeneralBus,
    ResponseMessageBus,
    ErrorMessageBus,
    NotificationMessageBus
}


export interface TransportMessage {
    id: string,
    dateCreated: Date,
    transport: Transport,
    target?: string,
    payload: any
}

export interface FisMessage {
    header: {
        messageID: string,
        messageName: `NotificationMessage` | `ResponseMessage` | `RequestMessage`
    },
    data: any
}


export interface Transmission {
    clientId: string,
    transmitter: MessageTransmitter,
    receiver: MessageReceiver,
    requestResponse: MessageRequestResponse,
    event: Observable<TransportEvent>
}

export type Transport = 'Websocket' | 'Http' | 'TCP'

export enum AdapterTransmissionRole {
    Transmitter,
    Receiver,
    RequestResponse
}



export type EventType = 'Server Started' | 'New Client' | 'Client Connected' | 'Client Re-connected' | 'Client Disconnected' | `Server Disconnected` | 'New Message' | `Notification` | `New Server` | `Server Connected` | `New Transport` | 'New Adapter' | 'Re-Flush'
export type ConnectionState = 'ONLINE' | 'OFFLINE'
export interface TransportService {
    getInfo(): Transport
    emit(message: TransportMessage): void
    subscribe(): Observable<TransportEvent> //all messages and whatever event will go through this, easier to implemnet across different transport protocol
}

export interface Info {
    transport: Transport
}

export interface ClientObject {
    id: string,
    dateCreated: Date,
    connectionState: BehaviorSubject<ConnectionState>
}
