import { Express } from 'express';
import { filter, Observable, Subject, Subscription, take } from "rxjs";
import { ClientObject, Transport, TransportEvent, TransportMessage, TransportService } from "../interface/connector.interface";
import { v4 as uuidv4 } from 'uuid'
import config from '../config/config.json';
import { handleClientHttpConnection, handleHttpClient, initiateClientToServer, startHttpServer } from "../utils/http.utils";
import { WrappedMessage } from '../utils/message.ordering';
import { error } from 'console';
import axios, { AxiosError } from 'axios';
import { EventMessage } from '../interface/transport.interface';
import ConsoleLogger from '../utils/log.utils';

export class HttpTransportService implements TransportService {
    private console: ConsoleLogger = new ConsoleLogger(`HttpTransportService`, ['transport'])
    private baseUrl!: string;
    private info: Transport = Transport.Http
    private connectedHttpServer: ConnectedHttpServer[] = [] // to allow the possibility of having to communicate with multiple servers as a client
    private connectedHttpClients: ConnectedHttpClient[] = [] // to keep track of the all the clients that are connected
    transportEvent!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>) {
        this.baseUrl = config.connection.transmitter
        this.transportEvent = event
    }

    public getInfo(): Transport {
        return this.info
    }

    public emit(message: TransportMessage): void {
        let clientObj: ConnectedHttpClient | undefined = this.connectedHttpClients.find(obj => obj.id == message.target)
        let serverObj: ConnectedHttpServer | undefined = this.connectedHttpServer.find(obj => obj.id === message.target)

        // for server usage
        if (clientObj && clientObj.connectionState.getValue() == 'ONLINE') {
            clientObj.responseStream.next(message.payload as WrappedMessage)
        }
        // for client usage 
        if (serverObj) {
            if (serverObj.connectionState.getValue() == 'ONLINE') {
                axios.post(`${this.baseUrl}message`, message.payload, {
                    headers: { 'Content-Type': 'application/json' },
                }).then((response) => {
                    this.console.log({ message: `Response From Server: ${response.data}`, details: response.status });
                }).catch((error: AxiosError) => {
                    console.error('HTTP emit error:', error.code);
                    this.transportEvent.next({
                        id: uuidv4(),
                        event: 'Re-Flush',
                        data: {
                            clientId: serverObj.id,
                            payload: message
                        } as EventMessage
                    } as TransportEvent)
                });
            } else {
                this.console.error({ message: `Target Server is offline: Reflusing message ${(message.payload as WrappedMessage).thisMessageID}` });
                this.transportEvent.next({
                    id: uuidv4(),
                    event: 'Re-Flush',
                    data: {
                        clientId: serverObj.id,
                        payload: message
                    } as EventMessage
                } as TransportEvent)
            }
        }

    }

    public subscribe(): Observable<TransportEvent> {
        return this.transportEvent.asObservable();
    }

    public startServer(port: number): void {
        startHttpServer(port).subscribe({
            next: (client: ConnectedHttpClient) => {
                handleHttpClient(client, this.connectedHttpClients).subscribe({
                    next: event => this.transportEvent.next(event),
                    error: error => console.error(error),
                    complete: () => (`Client ${client.id} disconnected...`)
                })
            },
            error: error => this.console.error({ message: 'Observer Error', details: error }),
            complete: () => this.console.log({ message: '...' })
        })
    }

    public startClient(url: string, receiverProfileInfo?: ConnectedHttpServer | undefined): void {
        initiateClientToServer(url, this.transportEvent, this.connectedHttpServer, receiverProfileInfo).then((connectedHttpServer: ConnectedHttpServer) => {
            handleClientHttpConnection(url, connectedHttpServer).subscribe({
                next: event => this.transportEvent.next(event),
                error: error => this.console.log({ message: `Observer Error`, details: error }),
                complete: () => {
                    this.console.log({ message: `Re-connecting with server` })
                    this.startClient(url, connectedHttpServer)
                }
            })
        }).catch((error: { error: AxiosError, objRef: ConnectedHttpServer | undefined }) => {
            this.console.error({ message: `HttpTransport ERROR: <InitiateClientToServerFailure>`, details: error.error.code })
            setTimeout(() => {
                this.startClient(url, error.objRef)
            }, 3000); // Retry with delay
        })
    }

}


export interface ConnectedHttpClient extends ClientObject {
    instance: Express
    responseStream: Subject<WrappedMessage>
}

export interface ConnectedHttpServer extends ClientObject {
}