import express, { Response } from 'express';
import * as fs from 'fs'
import { Express } from 'express';
import { v4 as uuidv4 } from 'uuid'
import { ConnectedHttpClient, ConnectedHttpServer } from "../transport/http";
import { BehaviorSubject, Observable, Observer, Subject, Subscription } from "rxjs";
import { ConnectionState, Transport, TransportEvent, TransportMessage } from '../interface/connector.interface';
import { EventMessage, FisMessage } from '../interface/transport.interface';
import { WrappedMessage } from './message.ordering';
import axios, { AxiosError, AxiosResponse } from 'axios';
import ConsoleLogger from './log.utils';
import path from 'path';
const console: ConsoleLogger = new ConsoleLogger(`HttpUtils`, ['transport'])

export function startHttpServer(port: number): Observable<ConnectedHttpClient> {
    return new Observable((observer: Observer<ConnectedHttpClient>) => {
        let app: Express = express();

        // Middleware to parse JSON requests
        app.use(express.json());

        app.listen(port, () => {
            console.log({ message: `Server running at http://localhost:${port}` });
        });

        observer.next({
            id: uuidv4(),
            dateCreated: new Date(),
            connectionState: new BehaviorSubject<ConnectionState>('ONLINE'),
            instance: app
        } as ConnectedHttpClient)
    })
}

export async function initiateClientToServer(url: string, event: Subject<TransportEvent>, connectedHttpServers: ConnectedHttpServer[], receiverProfileInfo: ConnectedHttpServer | undefined, browserEnv?: boolean,): Promise<ConnectedHttpServer> {
    return new Promise((resolve, reject) => {
        if (browserEnv) {
            // logic here for using browser fetch
        } else { // axios methods
            if (receiverProfileInfo) {
                console.log({ message: `Is Old profile, reconnecting with server` })
                checkOwnClientInfo(receiverProfileInfo.id).then((profile: ConnectedHttpServer) => {
                    receiverProfileInfo!.id = profile.id
                    // console.log({ message: 'jsonfile', details: profile })
                    postAxiosRequest(url + 'profile', { name: 'Old Client', message: profile }).then((profileInfo: { name: string, message: { id: string } }) => {
                        console.log({ message: `Acknowledged as previous client. Id: ${profileInfo.message.id}` })
                        event.next({
                            id: uuidv4(),
                            event: 'Server Connected',
                            data: {
                                clientId: profileInfo.message.id,
                                message: `Existing Http Channel ${profileInfo.message.id} re-established.`
                            } as EventMessage
                        })
                        // Update Http instance record
                        let clientObj: ConnectedHttpServer | undefined = connectedHttpServers.find(obj => obj.id === profileInfo.message.id)
                        console.log({ message: 'ClientObj', details: clientObj })
                        console.log({ message: 'ReceiverProfile', details: receiverProfileInfo })
                        if (clientObj) {
                            clientObj.connectionState.next('ONLINE')
                            console.log({ message: receiverProfileInfo.connectionState.getValue() })
                            resolve(clientObj)
                        }

                    }).catch((error: AxiosError) => {
                        reject({ error: error, objRef: receiverProfileInfo })
                    })
                }).catch((error) => {
                    console.error(error)
                    postAxiosRequest(url + 'profile', { name: 'New Client', data: null }).then((profileInfo: { name: string, message: any }) => {
                        updateProfileAndPublishEvent((receiverProfileInfo as ConnectedHttpServer), profileInfo, event, connectedHttpServers).then((receiverProfileInfo) => {
                            resolve(receiverProfileInfo)
                        })
                    }).catch((error) => {
                        reject({ error: error, objRef: receiverProfileInfo })
                    })
                    reject({ error: error, objRef: receiverProfileInfo })
                })
            } else {
                console.log({ message: `Is New profile, Connecting with server` })
                postAxiosRequest(url + 'profile', { name: 'New Client', data: null }).then((profileInfo: { name: string, message: any }) => {
                    updateProfileAndPublishEvent(receiverProfileInfo, profileInfo, event, connectedHttpServers).then((receiverProfileInfo) => {
                        resolve(receiverProfileInfo)
                    })
                }).catch((error) => {
                    reject({ error: error, objRef: receiverProfileInfo })
                })
            }
        }
    })
}

// For client usage
export function handleClientHttpConnection(url: string, server: ConnectedHttpServer): Observable<TransportEvent> {
    return new Observable((eventNotification: Observer<TransportEvent>) => {
        console.log({ message: `Long Poll Attempt for ${server.id}` })
        server.connectionState.next('ONLINE');
        let active: boolean = true; // Flag to control polling lifecycle

        const longPoll = async () => {
            while (active) {
                try {
                    // Axios request with timeout
                    const response = await axios.get(`${url}poll`); // removing the timeout temporarily. 
                    // const response = await axios.get(`${url}response`, {
                    //     timeout: 3000, // 10s timeout this one will trigger error. That's why it keeps on throwing error
                    // });

                    if (response.status === 200) {
                        const data = response.data;
                        console.log({ message: 'Long Poll Response', details: data })
                        eventNotification.next({
                            id: uuidv4(),
                            event: 'New Message',
                            data: {
                                id: uuidv4(),
                                dateCreated: new Date(),
                                transport: Transport.Http,
                                target: server.id,
                                payload: data,
                            } as TransportMessage,
                        });
                    } else if (response.status === 204) {
                        console.log({ message: 'No new messages from the server.' });
                    } else {
                        console.error({ message: `Unexpected response status: ${response.status}` })
                        handleServerConnectionError(active, eventNotification, server)
                        throw new Error(`Unexpected response status: ${response.status}`);
                    }
                } catch (error: unknown) {
                    console.error({ message: `Unknown Error.`, details: error }) // culprit is here
                    handleServerConnectionError(active, eventNotification, server)
                    // Error handling with server disconnect notification
                    let errorMessage: string;

                    if (axios.isAxiosError(error)) {
                        if (error.response) {
                            errorMessage = `Server returned status ${error.response.status}: ${error.response.statusText}`;
                        } else if (error.code === 'ECONNABORTED') {
                            errorMessage = 'Request timed out.';
                        } else {
                            errorMessage = error.message || 'An Axios error occurred.';
                        }
                    } else if (error instanceof Error) {
                        errorMessage = error.message;
                    } else {
                        errorMessage = 'An unknown error occurred during polling.';
                    }

                    console.error({ message: `Polling error: ${errorMessage}` });
                    // observer.error(new Error(errorMessage)); // Notify subscribers of the error
                    break; // Stop polling on error
                }
            }
        };

        longPoll();

        // Cleanup logic for unsubscribing
        return () => {
            console.log({ message: 'Unsubscribed from the long-polling channel.' });
            eventNotification.complete(); // Notify completion
        };
    });
}

function handleServerConnectionError(active: boolean, observer: Observer<TransportEvent>, server: ConnectedHttpServer): void {
    server.connectionState.next('OFFLINE');
    console.log({ message: 'Server lost connection' });
    active = false; // Stop polling
    observer.next({
        id: uuidv4(),
        event: 'Server Disconnected',
        data: {
            clientId: server.id,
            message: '',
            payload: {
                time: new Date(),
                objRef: server
            },
        } as EventMessage,
    });
    observer.complete()
}

async function updateProfileAndPublishEvent(receiverProfileInfo: ConnectedHttpServer | undefined, profile: { name: string, message: any }, event: Subject<TransportEvent>, connectedHttpServers: ConnectedHttpServer[]): Promise<ConnectedHttpServer> {
    return new Promise((resolve, reject) => {
        console.log({ message: `Assigned client Name: ${(profile.message as ConnectedHttpServer).id}` })
        receiverProfileInfo = profile.message as ConnectedHttpServer
        writeFile(profile.message as ConnectedHttpServer, (profile.message as ConnectedHttpServer).id).then(() => {
            event.next({
                id: uuidv4(),
                event: `New Server`,
                data: {
                    clientId: (profile.message as ConnectedHttpServer).id,
                    message: `New Http Channel ${(profile.message as ConnectedHttpServer).id} established.`
                } as EventMessage
            })
            // broadcast event to allow retransmission to relase buffered messages
            event.next({
                id: uuidv4(),
                event: `Server Connected`,
                data: {
                    clientId: (profile.message as ConnectedHttpServer).id,
                    message: `Server ${(profile.message as ConnectedHttpServer).id} connected and ready to go.`
                } as EventMessage
            })
        }).catch((error) => {
            reject(error)
        })

        // Update http instance record
        receiverProfileInfo = {
            id: (profile.message as ConnectedHttpServer).id,
            dateCreated: new Date(),
            connectionState: new BehaviorSubject<ConnectionState>(`ONLINE`)
        }
        connectedHttpServers.push(receiverProfileInfo)
        resolve(receiverProfileInfo)
    })
}

export async function postAxiosRequest(url: string, data: any): Promise<any> {
    return new Promise(async (resolve, reject) => {
        try {
            const response: AxiosResponse<any> = await axios.post(url, data);
            console.log({ message: 'Response', details: response.data });
            resolve(response.data)
        } catch (error) {
            if (axios.isAxiosError(error)) {
                console.error({ message: 'Axios Error:', details: error.code });
            } else {
                console.error({ message: 'Unexpected Error:', details: error });
            }
            reject(error)
        }
    })
}


export function handleHttpClient(clientInfo: ConnectedHttpClient, connectedClientHttp: ConnectedHttpClient[]): Observable<TransportEvent> {
    return new Observable((event: Observer<TransportEvent>) => {
        clientInfo.instance.post('/profile', (req, res) => {
            // Client will declare this first before attempting to poll for response channel
            handleProfile(clientInfo.instance, req.body, res, event, connectedClientHttp)
        });
    })
}

function handleProfile(app: Express, data: { name: `Old Client` | `New Client`, message: any }, res: Response, event: Observer<TransportEvent>, connectedClientHttp: ConnectedHttpClient[]): void {
    if (data.name == `New Client`) {
        let clientInstance: ConnectedHttpClient = {
            id: uuidv4(), // client should only be assigned at this level. And is passed around for reference pointing
            dateCreated: new Date(),
            instance: app,
            connectionState: new BehaviorSubject<ConnectionState>(`OFFLINE`), // for now it's offline because it needs to establish the long polling first
            responseStream: new Subject<WrappedMessage>()
        }

        // send to receiver for reference
        res.json({
            name: `New Profile`, message: { id: clientInstance.id }
        })
        // publish first event notification
        event.next({
            id: uuidv4(),
            event: `New Client`,
            data: {
                clientId: clientInstance.id,
                message: `New Http Client Connected. Adapter ID assigned: ${clientInstance.id}`,
                payload: clientInstance
            } as EventMessage
        })
        // Update connected clientInstance info to adapter
        connectedClientHttp.push(clientInstance)
        addClientToDB(clientInstance)
        startListeningAndStreaming(app, clientInstance, event)
    } else if (data.name == 'Old Client') {
        console.log({ message: `Is old client`, details: data })
        // update first
        let clientInstance: ConnectedHttpClient | undefined
        if (connectedClientHttp.length > 0) {
            clientInstance = connectedClientHttp.find(obj => obj.id === data.message.id)
            handleFoundClient(clientInstance)
        } else {
            // for the case server itself got shit down or something
            checkIfClientExists(data.message.id).then((client: ConnectedHttpClient) => {
                clientInstance = client
                handleFoundClient(clientInstance)
            }).catch(error => console.error(error))
        }
        function handleFoundClient(clientInstance: ConnectedHttpClient | undefined): void {
            if (clientInstance) {
                console.log({ message: `Http Client ${clientInstance.id} Found` })
                res.json({ name: 'Adjusted Profile', message: { id: clientInstance.id } })
                // replace socket instance since the previous has been terminated
                clientInstance.instance = app
                // some explanation here. For the case where the server reads from the DB, no need to terminate subject, since all instances would be destroyed alongside the server shut down. This case is specificd only when there's a need to read from local file
                if (!clientInstance.connectionState) {
                    clientInstance.connectionState = new BehaviorSubject<ConnectionState>(`OFFLINE`)
                }
                // need to start listening again, because it's assigned a different socket instance this time round
                startListeningAndStreaming(app, clientInstance, event, true)
                event.next({
                    id: uuidv4(),
                    event: 'Client Connected',
                    data: {
                        clientId: clientInstance.id,
                        message: `Client ${clientInstance.id} connection re-established`,
                        payload: clientInstance
                    } as EventMessage
                })

            } else {
                console.log({ message: `Profile Not Found` })
                res.json({ name: 'Error', message: 'Receiver Profile Not found' })
            }
        }
    }
}

export async function checkIfClientExists(id: string): Promise<ConnectedHttpClient> {
    return new Promise((resolve, reject) => {
        try {
            // Check if the file exists
            let filePath = process.env.FolderPath as string + 'clients.json'
            if (!fs.existsSync(filePath)) {
                console.log({ message: `File does not exist.` })
                reject('File does not exist');
            }

            // Read and parse the data
            const fileContent = fs.readFileSync(filePath, 'utf-8');
            const data: any[] = JSON.parse(fileContent);

            // Check if an details with the given id exists
            let obj = data.find(entry => entry.id === id);

            if (obj) {
                console.log({ message: `Client with ID ${id} exists.` })
            } else {
                console.log({ message: `Client with ID ${id} does not exist.` })
            }
            resolve(obj);
        } catch (error) {
            reject(`Error reading the file`)
        }
    })
}

/* For Internal Usage only. Temporary serve as a way for server to keep track of clients. To be replaced in the future with better alternatives. */
export function addClientToDB(entry: ConnectedHttpClient): void {
    try {
        let data: ConnectedHttpClient[] = [];
        let filePath = process.env.FolderPath as string + 'clients.json'
        // Check if the file exists and load existing data
        if (fs.existsSync(filePath)) {
            const fileContent = fs.readFileSync(filePath, 'utf-8');
            data = JSON.parse(fileContent);
        }

        // Append the new object to the array
        data.push({
            id: entry.id,
            dateCreated: entry.dateCreated,
            connectionState: null,
            instance: null
        } as unknown as ConnectedHttpClient);

        // Write the updated array back to the file
        fs.writeFileSync(filePath, JSON.stringify(data, null, 2), 'utf-8');
        console.log({ message: `Entry added successfully.` });
    } catch (error) {
        console.error({ message: 'Error writing to file:', details: error });
    }
}

// this is for server usage only
export function startListeningAndStreaming(app: Express, client: ConnectedHttpClient, eventListener: Observer<TransportEvent>, oldClient?: boolean): void {
    /* Generally, we don't need this unless in the case of being the receiver */
    app.post('/message', (req, res) => {
        eventListener.next({
            id: uuidv4(),
            event: 'New Message',
            data: {
                id: uuidv4(),
                dateCreated: new Date(),
                transport: Transport.Http,
                target: client.id, // this ref to be associated with the client/channel
                payload: req.body
            } as TransportMessage
        })
        res.json(`Received ${((req.body as WrappedMessage)?.payload as FisMessage)?.header?.messageID ?? `Undefined`}`)
    })

    app.get('/poll', (req, res) => {
        console.log({ message: 'Client connected for long polling.' });
        client.connectionState.next('ONLINE');

        // notify it's associated retransmission to start releaseing buffer
        eventListener.next({
            id: uuidv4(),
            event: oldClient ? 'Client Re-connected' : `Client Connected`,
            data: {
                clientId: client.id,
                message: `Socket Client ${oldClient ? `Re-Connected` : `Connected`}. Adapter ID assigned: ${client.id}`,
                payload: client
            } as EventMessage
        })

        // Flag to track if the response has been sent
        let responseSent = false;

        // Subscribe to the data stream
        const subscription = client.responseStream.asObservable().subscribe({
            next: (message: WrappedMessage) => {
                if (!responseSent) {
                    console.log({ message: `Sending data ${message.thisMessageID} to client ${client.id}}` });
                    res.json(message); // Send the data to the client
                    responseSent = true; // Mark response as sent
                    subscription.unsubscribe(); // Unsubscribe to close this request
                }
            },
            error: (err) => {
                if (!responseSent) {
                    console.error({ message: 'Error in data stream:', details: err });
                    res.status(500).send('Internal Server Error');
                    responseSent = true; // Mark response as sent
                }
                subscription.unsubscribe(); // Ensure cleanup
            },
            complete: () => {
                if (!responseSent) {
                    console.log({ message: 'Data stream completed.' });
                    res.status(204).send(); // No Content
                    responseSent = true; // Mark response as sent
                }
                subscription.unsubscribe(); // Ensure cleanup
            },
        });

        // Timeout if no data is emitted within a specified duration
        const timeout = setTimeout(() => {
            if (!responseSent) {
                console.log({ message: 'No data emitted. Sending timeout response.' });
                res.status(204).send(); // No Content
                responseSent = true; // Mark response as sent
                subscription.unsubscribe(); // Ensure cleanup
            }
        }, 15000); // 15 seconds timeout (adjust as needed)

        // Handle client disconnection
        res.on('close', () => {
            if (!responseSent) {
                console.error({ message: `Http Client ${client.id} disconnected` });
                eventListener.next({
                    id: uuidv4(),
                    event: 'Client Disconnected',
                    data: {
                        clientId: client.id,
                        payload: {
                            time: new Date()
                        }
                    } as EventMessage
                } as TransportEvent)
                client.connectionState.next(`OFFLINE`)
                subscription.unsubscribe(); // Ensure cleanup
            }
            clearTimeout(timeout); // Clear timeout to avoid unnecessary execution
        });
    });

}


// Check if filename exists. Return profile information if there's any
export async function checkOwnClientInfo(filename?: string): Promise<ConnectedHttpServer> {
    return new Promise((resolve, reject) => {
        // Check if the file exists
        let filePath = process.env.FolderPath as string
        if (fs.existsSync(filePath + `${filename}.json`)) {
            try {
                // Read the file contents
                const fileData = fs.readFileSync(`${filename}.json`, 'utf8');

                // If the file is empty, return an error
                if (fileData.trim() === "") {
                    throw new Error("File is empty");
                }

                // Parse and return the data if present
                const jsonData = JSON.parse(fileData);
                resolve(jsonData)

            } catch (err) {
                // Handle parsing errors or other file-related errors
                let errMsg: string = ("Error reading or parsing file: " + err)
                reject(errMsg);
            }
        } else {
            reject("File does not exist");
        }
    })
}

// Specifically to write receiver profile information
async function writeFile(data: ConnectedHttpServer, filename: string): Promise<boolean> {
    return new Promise((resolve, reject) => {
        // Ensure the folder exists
        const folderPath = process.env.FolderPath as string
        console.log({ message: folderPath })
        // const folderPath = path.join(__dirname, folder);
        if (!fs.existsSync(folderPath)) {
            fs.mkdirSync(folderPath, { recursive: true }); // Create folder if it doesn't exist
        } else {
            console.log({ message: 'Folder already exist' })
        }

        // Construct the full file path (include the folder)
        const filePath = path.join(folderPath, `${filename}.json`);

        // Write JSON data to a file
        fs.writeFile(filePath, JSON.stringify(data, null, 2), (err) => {
            if (err) {
                console.log({ message: 'Error writing file', details: err });
                reject(false);
            } else {
                console.log({ message: 'File has been written', details: filePath });
                resolve(true);
            }
        });
    });
}

